"""
    conftest for mutacc-auto api
"""

import pytest

from cg.apps.mutacc_auto import MutaccAutoAPI

CONFIG = {
    "mutacc-auto": {
        "config_path": "mutacc-auto_config",
        "binary_path": "mutacc-auto",
        "padding": 111,
    }
}


@pytest.fixture(scope="function")
def mutacc_config():
    """
    mutacc config fixture
    """

    _config = CONFIG
    return _config


@pytest.fixture(scope="function")
def mutacc_auto_api():
    """
    mutacc-auto api
    """

    _api = MutaccAutoAPI(CONFIG)

    return _api


class MockFailedProcess:
    """
    Mock a failed process from subprocess.run()
    """

    @property
    def returncode(self):
        """ Mock returncode that is not 0"""
        return 1


@pytest.fixture(scope="function")
def mock_failed_process():
    """Return a completed process from subprocess.run()"""
    return MockFailedProcess()



"""Set data in the status database and LIMS"""
import datetime
import getpass

import click
from cg.apps.lims import LimsAPI
from .families import families
from .family import family
from cg.constants import FAMILY_ACTIONS, PRIORITY_OPTIONS, FLOWCELL_STATUS
from cg.exc import LimsDataError
from cg.store import Store, models

CONFIRM = "Continue?"
HELP_KEY_VALUE = "Give a property on sample and the value to set it to, e.g. -kv name Prov52"
HELP_SKIP_LIMS = "Skip setting value in LIMS"
HELP_YES = "Answer yes on all confirmations"
OPTION_LONG_KEY_VALUE = "--key-value"
OPTION_LONG_SKIP_LIMS = "--skip-lims"
OPTION_LONG_YES = "--yes"
OPTION_SHORT_KEY_VALUE = "-kv"
OPTION_SHORT_YES = "-y"
NOT_CHANGABLE_SAMPLE_ATTRIBUTES = [
    "application_version_id",
    "customer_id",
    "deliveries",
    "flowcells",
    "id",
    "internal_id",
    "invoice",
    "invoice_id",
    "is_external",
    "links",
    "state",
    "to_dict",
]


@click.group("set")
@click.pass_context
def set_cmd(context):
    """Update information in the database."""
    context.obj["status_db"] = Store(context.obj["database"])
    context.obj["lims_api"] = LimsAPI(context.obj)


@set_cmd.command()
@click.option(
    "-id",
    "--identifier",
    "identifiers",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help="Give an identifier on sample and the value to use it with, e.g. -id name Prov52",
)
@click.option(
    "-kv",
    "--key-value",
    "kwargs",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help="Give a property on sample and the value to set it to, e.g. -kv name Prov52",
)
@click.option("--skip-lims", is_flag=True, help="Skip setting value in LIMS")
@click.option("-y", "--yes", is_flag=True, help="Answer yes on all confirmations")
@click.argument("case_id", required=False)
@click.pass_context
def samples(
    context: click.Context,
    identifiers: click.Tuple([str, str]),
    kwargs: click.Tuple([str, str]),
    skip_lims: bool,
    yes: bool,
    case_id: str,
):
    """Set values on many samples at the same time"""
    store = context.obj["status_db"]
    sample_objs = _get_samples(case_id, identifiers, store)

    if not sample_objs:
        click.echo(click.style(fg="red", text="No samples to alter!"))
        context.abort()

    click.echo("Would alter samples:")

    for sample_obj in sample_objs:
        click.echo(f"{sample_obj}")

    if not (yes or click.confirm(CONFIRM)):
        context.abort()

    for sample_obj in sample_objs:
        context.invoke(
            sample, sample_id=sample_obj.internal_id, kwargs=kwargs, yes=yes, skip_lims=skip_lims
        )


def _get_samples(
    case_id: str, identifiers: click.Tuple([str, str]), store: Store
) -> [models.Sample]:
    """Get samples that match both case_id and identifiers if given"""
    samples_by_case_id = None
    samples_by_id = None

    if case_id:
        samples_by_case_id = _get_samples_by_case_id(case_id, store)

    if identifiers:
        samples_by_id = _get_samples_by_identifiers(identifiers, store)

    if case_id and identifiers:
        sample_objs = set(set(samples_by_case_id) & set(samples_by_id))
    else:
        sample_objs = samples_by_case_id or samples_by_id

    return sample_objs


def _get_samples_by_identifiers(
    identifiers: click.Tuple([str, str]), store: Store
) -> models.Sample:
    """Get samples matched by given set of identifiers"""
    identifier_args = {}
    for identifier_name, identifier_value in identifiers:
        identifier_args[identifier_name] = identifier_value
    return store.samples_by_ids(**identifier_args)


def _get_samples_by_case_id(case_id: str, store: Store) -> [models.Sample]:
    """Get samples on a given case-id"""
    case = store.family(internal_id=case_id)
    return [link.sample for link in case.links] if case else []


def is_locked_attribute_on_sample(key, skip_attributes):
    """Returns true if the attribute is private or in the skip list"""
    return is_private_attribute(key) or key in skip_attributes


def is_private_attribute(key):
    """Returns true if key has name indicating it is private"""
    return key.startswith("_")


def list_changeable_sample_attributes(sample_obj: models.Sample = None, skip_attributes: [] = None):
    """List changeable attributes on sample and its current value"""

    sample_attributes = models.Sample.__dict__.keys()
    for attribute in sample_attributes:
        if is_locked_attribute_on_sample(attribute, skip_attributes):
            continue

        message = attribute

        if sample_obj:
            message += f": {sample_obj.__dict__.get(attribute)}"

        click.echo(message)


def show_set_sample_help(sample_obj: models.Sample = "None") -> None:
    """Show help for the set sample command"""
    click.echo("sample_id: optional, internal_id of sample to set value on")
    show_option_help(long_name=OPTION_LONG_SKIP_LIMS, help_text=HELP_SKIP_LIMS)
    show_option_help(short_name=OPTION_SHORT_YES, long_name=OPTION_LONG_YES, help_text=HELP_YES)
    show_option_help(
        short_name=OPTION_SHORT_KEY_VALUE, long_name=OPTION_LONG_KEY_VALUE, help_text=HELP_KEY_VALUE
    )
    list_changeable_sample_attributes(sample_obj, skip_attributes=NOT_CHANGABLE_SAMPLE_ATTRIBUTES)
    click.echo(f"To set apptag use '{OPTION_SHORT_KEY_VALUE} application_version [APPTAG]")
    click.echo(f"To set customer use '{OPTION_SHORT_KEY_VALUE} customer [CUSTOMER]")


def show_option_help(short_name: str = None, long_name: str = None, help_text: str = None):
    """Show help for one option"""
    help_message = f"Use "

    if short_name:
        help_message += f"'{short_name}'"

    if short_name and long_name:
        help_message += " or "

    if long_name:
        help_message += f"'{long_name}'"

    if help_text:
        help_message += f": {help_text}"

    click.echo(help_message)


@set_cmd.command()
@click.argument("sample_id", required=False)
@click.option(
    OPTION_SHORT_KEY_VALUE,
    OPTION_LONG_KEY_VALUE,
    "kwargs",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help=HELP_KEY_VALUE,
)
@click.option(OPTION_LONG_SKIP_LIMS, is_flag=True, help=HELP_SKIP_LIMS)
@click.option(OPTION_SHORT_YES, OPTION_LONG_YES, is_flag=True, help=HELP_YES)
@click.option("--help", is_flag=True)
@click.pass_context
def sample(context, sample_id, kwargs, skip_lims, yes, help):
    sample_obj = context.obj["status_db"].sample(internal_id=sample_id)

    if help:
        show_set_sample_help(sample_obj)

    if sample_obj is None:
        click.echo(click.style(f"Can't find sample {sample_id}", fg="red"))
        context.abort()

    for key, value in kwargs:

        if is_locked_attribute_on_sample(key, NOT_CHANGABLE_SAMPLE_ATTRIBUTES):
            click.echo(click.style(f"{key} is not a changeable attribute on sample", fg="yellow"))
            continue
        if not hasattr(sample_obj, key):
            click.echo(click.style(f"{key} is not a property of sample", fg="yellow"))
            continue
        new_value = None
        if key in ["customer", "application_version"]:
            if key == "customer":
                new_value = context.obj["status_db"].customer(value)
            elif key == "application_version":
                new_value = context.obj["status_db"].current_application_version(value)

            if not new_value:
                click.echo(click.style(f"{key} {value} not found, aborting", fg="red"))
                context.abort()
        else:
            new_value = value

        old_value = getattr(sample_obj, key)

        click.echo(f"Would change from {key}={old_value} to {key}={new_value} on {sample_obj}")

        if not (yes or click.confirm(CONFIRM)):
            continue

        setattr(sample_obj, key, new_value)
        _update_comment(_generate_comment(key, old_value, new_value), sample_obj)
        context.obj["status_db"].commit()

    if not skip_lims:

        for key, value in kwargs:
            click.echo(f"Would set {key} to {value} for {sample_obj.internal_id} in LIMS")

            if not (yes or click.confirm(CONFIRM)):
                context.abort()

            try:
                context.obj["lims_api"].update_sample(lims_id=sample_id, **{key: value})
                click.echo(click.style(f"Set LIMS/{key} to {value}", fg="blue"))
            except LimsDataError as err:
                click.echo(
                    click.style(f"Failed to set LIMS/{key} to {value}, {err.message}", fg="red")
                )


def _generate_comment(what, old_value, new_value):
    """Generate a comment that can be used in the comment field to describe updated value"""
    return f"\n{what} changed from {str(old_value)} to {str(new_value)}."


def _update_comment(comment, obj):
    """Appends the comment on obj including a timestamp"""
    if comment:
        timestamp = str(datetime.datetime.now())[:-10]
        if obj.comment is None:
            obj.comment = f"{timestamp}-{getpass.getuser()}: {comment}"
        else:
            obj.comment = f"{timestamp}-{getpass.getuser()}: {comment}" + "\n" + obj.comment


@set_cmd.command()
@click.option("-s", "--status", type=click.Choice(FLOWCELL_STATUS))
@click.argument("flowcell_name")
@click.pass_context
def flowcell(context, flowcell_name, status):
    """Update information about a flowcell"""
    flowcell_obj = context.obj["status_db"].flowcell(flowcell_name)

    if flowcell_obj is None:
        click.echo(click.style(f"flowcell not found: {flowcell_name}", fg="yellow"))
        context.abort()
    prev_status = flowcell_obj.status
    flowcell_obj.status = status

    context.obj["status_db"].commit()
    click.echo(click.style(f"{flowcell_name} set: {prev_status} -> {status}", fg="green"))


set_cmd.add_command(family)
set_cmd.add_command(families)



import datetime as dt

import pytest
from cg.constants import Pipeline
from cg.exc import OrderError
from cg.meta.orders.status import StatusHandler


def test_pools_to_status(rml_order_to_submit):
    # GIVEN a rml order with three samples in one pool
    # WHEN parsing for status
    data = StatusHandler.pools_to_status(rml_order_to_submit)
    # THEN it should pick out the general information
    assert data["customer"] == "cust001"
    assert data["order"] == "ctDNA sequencing - order 9"
    # ... and information about the pool(s)
    assert len(data["pools"]) == 1
    assert data["pools"][0]["name"] == "pool-1"
    assert data["pools"][0]["application"] == "RMLS05R150"
    assert data["pools"][0]["data_analysis"] == str(Pipeline.FASTQ)
    assert data["pools"][0]["capture_kit"] == "Agilent Sureselect CRE"


def test_samples_to_status(fastq_order_to_submit):
    # GIVEN fastq order with two samples
    # WHEN parsing for status
    data = StatusHandler.samples_to_status(fastq_order_to_submit)
    # THEN it should pick out samples and relevant information
    assert len(data["samples"]) == 2
    first_sample = data["samples"][0]
    assert first_sample["name"] == "prov1"
    assert first_sample["application"] == "WGSPCFC060"
    assert first_sample["priority"] == "priority"
    assert first_sample["tumour"] is False

    # ... and the other sample is a tumour
    assert data["samples"][1]["tumour"] is True


def test_microbial_samples_to_status(microbial_order_to_submit):
    # GIVEN microbial order with three samples

    # WHEN parsing for status
    data = StatusHandler.microbial_samples_to_status(microbial_order_to_submit)

    # THEN it should pick out samples and relevant information
    assert len(data["samples"]) == 5
    assert data["customer"] == "cust002"
    assert data["order"] == "Microbial samples"
    assert data["comment"] == "Order comment"

    # THEN first sample should contain all the relevant data from the microbial order
    sample_data = data["samples"][0]
    assert sample_data.get("priority") in "research"
    assert sample_data["name"] == "all-fields"
    assert sample_data.get("internal_id") is None
    assert sample_data["organism_id"] == "M.upium"
    assert sample_data["reference_genome"] == "NC_111"
    assert sample_data["application"] == "MWRNXTR003"
    assert sample_data["comment"] == "plate comment"


def test_families_to_status(mip_order_to_submit):
    # GIVEN a scout order with a trio family
    # WHEN parsing for status
    data = StatusHandler.cases_to_status(mip_order_to_submit)
    # THEN it should pick out the family
    assert len(data["families"]) == 2
    family = data["families"][0]
    assert family["name"] == "family1"
    assert family["data_analysis"] == str(Pipeline.MIP_DNA)
    assert family["priority"] == "standard"
    assert set(family["panels"]) == {"IEM"}
    assert len(family["samples"]) == 3

    first_sample = family["samples"][0]
    assert first_sample["name"] == "sample1"
    assert first_sample["application"] == "WGTPCFC030"
    assert first_sample["sex"] == "female"
    assert first_sample["status"] == "affected"
    assert first_sample["mother"] == "sample2"
    assert first_sample["father"] == "sample3"

    # ... second sample has a comment
    assert isinstance(family["samples"][1]["comment"], str)


def test_store_rml(orders_api, base_store, rml_status_data):

    # GIVEN a basic store with no samples and a rml order
    assert base_store.pools(customer=None).count() == 0
    # WHEN storing the order
    new_pools = orders_api.store_pools(
        customer=rml_status_data["customer"],
        order=rml_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=1234348,
        pools=rml_status_data["pools"],
    )
    # THEN it should update the database with new pools
    assert len(new_pools) == 1
    assert base_store.pools(customer=None).count() == 1
    new_pool = base_store.pools(customer=None).first()
    assert new_pool == new_pools[0]
    assert new_pool.name == "pool-1"
    assert new_pool.application_version.application.tag == "RMLS05R150"
    assert new_pool.data_analysis == str(Pipeline.FASTQ)
    assert new_pool.capture_kit == "Agilent Sureselect CRE"
    # ... and add a delivery
    assert len(new_pool.deliveries) == 1
    assert new_pool.deliveries[0].destination == "caesar"


def test_store_rml_bad_apptag(orders_api, base_store, rml_status_data):

    # GIVEN a basic store with no samples and a rml order
    assert base_store.pools(customer=None).count() == 0

    for pool in rml_status_data["pools"]:
        pool["application"] = "nonexistingtag"

    # THEN it should raise OrderError
    with pytest.raises(OrderError):
        # WHEN storing the order
        orders_api.store_pools(
            customer=rml_status_data["customer"],
            order=rml_status_data["order"],
            ordered=dt.datetime.now(),
            ticket=1234348,
            pools=rml_status_data["pools"],
        )


def test_store_samples(orders_api, base_store, fastq_status_data):
    # GIVEN a basic store with no samples and a fastq order
    assert base_store.samples().count() == 0
    assert base_store.families().count() == 0

    # WHEN storing the order
    new_samples = orders_api.store_fastq_samples(
        customer=fastq_status_data["customer"],
        order=fastq_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=1234348,
        samples=fastq_status_data["samples"],
    )

    # THEN it should store the samples and create a family for each sample
    assert len(new_samples) == 2
    assert base_store.samples().count() == 2
    assert base_store.families().count() == 2
    first_sample = new_samples[0]
    assert len(first_sample.links) == 1
    family_link = first_sample.links[0]
    assert family_link.family in base_store.families()
    for sample in new_samples:
        assert len(sample.deliveries) == 1


def test_store_samples_data_analysis_stored(orders_api, base_store, fastq_status_data):
    # GIVEN a basic store with no samples and a fastq order
    assert base_store.samples().count() == 0
    assert base_store.families().count() == 0

    # WHEN storing the order
    new_samples = orders_api.store_fastq_samples(
        customer=fastq_status_data["customer"],
        order=fastq_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=1234348,
        samples=fastq_status_data["samples"],
    )

    # THEN the sample sex should be stored
    assert new_samples[0].sex == "male"


def test_store_samples_bad_apptag(orders_api, base_store, fastq_status_data):
    # GIVEN a basic store with no samples and a fastq order
    assert base_store.samples().count() == 0
    assert base_store.families().count() == 0

    for sample in fastq_status_data["samples"]:
        sample["application"] = "nonexistingtag"

    # THEN it should raise OrderError
    with pytest.raises(OrderError):
        # WHEN storing the order
        orders_api.store_fastq_samples(
            customer=fastq_status_data["customer"],
            order=fastq_status_data["order"],
            ordered=dt.datetime.now(),
            ticket=1234348,
            samples=fastq_status_data["samples"],
        )


def test_store_microbial_samples(orders_api, base_store, microbial_status_data):

    # GIVEN a basic store with no samples and a microbial order and one Organism
    assert base_store.samples().count() == 0
    assert base_store.families().count() == 0
    assert base_store.organisms().count() == 1

    # WHEN storing the order
    new_samples = orders_api.store_microbial_samples(
        customer=microbial_status_data["customer"],
        order=microbial_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=1234348,
        samples=microbial_status_data["samples"],
        comment="",
        data_analysis=Pipeline.MICROSALT,
    )

    # THEN it should store the samples under a case (family) and the used previously unknown
    # organisms
    assert new_samples
    assert base_store.families().count() == 1
    assert len(new_samples) == 5
    assert base_store.samples().count() == 5
    assert base_store.organisms().count() == 3


def test_store_microbial_case_data_analysis_stored(orders_api, base_store, microbial_status_data):

    # GIVEN a basic store with no samples and a microbial order and one Organism
    assert base_store.samples().count() == 0
    assert base_store.families().count() == 0

    # WHEN storing the order
    new_samples = orders_api.store_microbial_samples(
        customer=microbial_status_data["customer"],
        order=microbial_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=1234348,
        samples=microbial_status_data["samples"],
        comment="",
        data_analysis=Pipeline.MICROSALT,
    )

    # THEN store the samples under a case with the microbial data_analysis type on case level
    assert base_store.samples().count() > 0
    assert base_store.families().count() == 1

    microbial_case = base_store.families().first()
    assert microbial_case.data_analysis == str(Pipeline.MICROSALT)


def test_store_microbial_samples_bad_apptag(orders_api, microbial_status_data):

    # GIVEN a basic store missing the application on the samples

    for sample in microbial_status_data["samples"]:
        sample["application"] = "nonexistingtag"

    # THEN it should raise OrderError
    with pytest.raises(OrderError):
        # WHEN storing the order
        orders_api.store_microbial_samples(
            customer=microbial_status_data["customer"],
            order=microbial_status_data["order"],
            ordered=dt.datetime.now(),
            ticket=1234348,
            samples=microbial_status_data["samples"],
            comment="",
            data_analysis=Pipeline.MICROSALT,
        )


def test_store_microbial_sample_priority(orders_api, base_store, microbial_status_data):

    # GIVEN a basic store with no samples
    assert base_store.samples().count() == 0

    # WHEN storing the order
    orders_api.store_microbial_samples(
        customer=microbial_status_data["customer"],
        order=microbial_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=1234348,
        samples=microbial_status_data["samples"],
        comment="",
        data_analysis=Pipeline.MICROSALT,
    )

    # THEN it should store the sample priority
    assert base_store.samples().count() > 0
    microbial_sample = base_store.samples().first()

    assert microbial_sample.priority_human == "research"


def test_store_mip(orders_api, base_store, mip_status_data):
    # GIVEN a basic store with no samples or nothing in it + scout order
    assert base_store.samples().first() is None
    assert base_store.families().first() is None

    # WHEN storing the order
    new_families = orders_api.store_cases(
        customer=mip_status_data["customer"],
        order=mip_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=123456,
        cases=mip_status_data["families"],
    )

    # THEN it should create and link samples and the family
    assert len(new_families) == 2
    new_family = new_families[0]
    assert new_family.name == "family1"
    assert set(new_family.panels) == {"IEM"}
    assert new_family.priority_human == "standard"

    assert len(new_family.links) == 3
    new_link = new_family.links[0]
    assert new_family.data_analysis == str(Pipeline.MIP_DNA)
    assert new_link.status == "affected"
    assert new_link.mother.name == "sample2"
    assert new_link.father.name == "sample3"
    assert new_link.sample.name == "sample1"
    assert new_link.sample.sex == "female"
    assert new_link.sample.application_version.application.tag == "WGTPCFC030"
    assert new_link.sample.is_tumour
    assert isinstance(new_family.links[1].sample.comment, str)

    assert base_store.deliveries().count() == base_store.samples().count()
    for link in new_family.links:
        assert len(link.sample.deliveries) == 1


def test_store_mip_rna(orders_api, base_store, mip_rna_status_data):
    # GIVEN a basic store with no samples or nothing in it + rna order
    rna_application = "RNAPOAR025"
    assert base_store.samples().first() is None
    assert base_store.families().first() is None
    assert base_store.application(rna_application)

    # WHEN storing the order
    new_cases = orders_api.store_cases(
        customer=mip_rna_status_data["customer"],
        order=mip_rna_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=123456,
        cases=mip_rna_status_data["families"],
    )

    # THEN it should create and link samples and the casing
    assert len(new_cases) == 1
    new_casing = new_cases[0]

    assert len(new_casing.links) == 2
    new_link = new_casing.links[0]
    assert new_casing.data_analysis == str(Pipeline.MIP_RNA)
    assert new_link.sample.name == "sample1-rna-t1"
    assert new_link.sample.application_version.application.tag == rna_application
    assert new_link.sample.time_point == 1
    assert new_link.sample.from_sample == "sample1"


def test_store_families_bad_apptag(orders_api, base_store, mip_status_data):
    # GIVEN a basic store with no samples or nothing in it + scout order
    assert base_store.samples().first() is None
    assert base_store.families().first() is None

    for family in mip_status_data["families"]:
        for sample in family["samples"]:
            sample["application"] = "nonexistingtag"

    # THEN it should raise OrderError
    with pytest.raises(OrderError):
        # WHEN storing the order
        orders_api.store_cases(
            customer=mip_status_data["customer"],
            order=mip_status_data["order"],
            ordered=dt.datetime.now(),
            ticket=123456,
            cases=mip_status_data["families"],
        )


def test_store_external(orders_api, base_store, external_status_data):

    # GIVEN a basic store with no samples or nothing in it + external order
    assert base_store.samples().first() is None
    assert base_store.families().first() is None

    # WHEN storing the order
    new_families = orders_api.store_cases(
        customer=external_status_data["customer"],
        order=external_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=123456,
        cases=external_status_data["families"],
    )

    # THEN it should create and link samples and the family
    family_obj = base_store.families().first()
    assert len(new_families) == 1
    new_family = new_families[0]
    assert new_family == family_obj
    assert new_family.name == "fam2"
    assert new_family.data_analysis == str(Pipeline.MIP_DNA)
    assert set(new_family.panels) == {"CTD", "CILM"}
    assert new_family.priority_human == "priority"

    assert len(new_family.links) == 2
    new_link = new_family.links[0]
    assert new_link.status == "affected"
    assert new_link.sample.name == "sample1"
    assert new_link.sample.sex == "male"
    assert new_link.sample.capture_kit == "Agilent Sureselect V5"
    assert new_link.sample.application_version.application.tag == "EXXCUSR000"
    assert isinstance(new_family.links[0].sample.comment, str)

    assert base_store.deliveries().count() == base_store.samples().count()
    for link in new_family.links:
        assert len(link.sample.deliveries) == 1


def test_store_external_bad_apptag(orders_api, base_store, external_status_data):

    # GIVEN a basic store with no samples or nothing in it + external order
    assert base_store.samples().first() is None
    assert base_store.families().first() is None

    for family in external_status_data["families"]:
        for sample in family["samples"]:
            sample["application"] = "nonexistingtag"

    # THEN it should raise OrderError
    with pytest.raises(OrderError):
        # WHEN storing the order
        orders_api.store_cases(
            customer=external_status_data["customer"],
            order=external_status_data["order"],
            ordered=dt.datetime.now(),
            ticket=123456,
            cases=external_status_data["families"],
        )


def test_store_metagenome_samples(orders_api, base_store, metagenome_status_data):

    # GIVEN a basic store with no samples and a metagenome order
    assert base_store.samples().count() == 0

    # WHEN storing the order
    new_samples = orders_api.store_samples(
        customer=metagenome_status_data["customer"],
        order=metagenome_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=1234348,
        samples=metagenome_status_data["samples"],
    )

    # THEN it should store the samples
    assert len(new_samples) == 2
    assert base_store.samples().count() == 2


def test_store_metagenome_samples(orders_api, base_store, metagenome_status_data):

    # GIVEN a basic store with no samples and a metagenome order
    assert base_store.samples().count() == 0

    # WHEN storing the order
    new_samples = orders_api.store_samples(
        customer=metagenome_status_data["customer"],
        order=metagenome_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=1234348,
        samples=metagenome_status_data["samples"],
    )

    # THEN it should have stored the samples
    assert base_store.samples().count() > 0


def test_store_metagenome_samples_bad_apptag(orders_api, base_store, metagenome_status_data):

    # GIVEN a basic store with no samples and a metagenome order
    assert base_store.samples().count() == 0

    for sample in metagenome_status_data["samples"]:
        sample["application"] = "nonexistingtag"

    # THEN it should raise OrderError
    with pytest.raises(OrderError):
        # WHEN storing the order
        orders_api.store_samples(
            customer=metagenome_status_data["customer"],
            order=metagenome_status_data["order"],
            ordered=dt.datetime.now(),
            ticket=1234348,
            samples=metagenome_status_data["samples"],
        )


def test_store_cancer_samples(orders_api, base_store, balsamic_status_data):

    # GIVEN a basic store with no samples and a cancer order
    assert base_store.samples().first() is None
    assert base_store.families().first() is None

    # WHEN storing the order
    new_families = orders_api.store_cases(
        customer=balsamic_status_data["customer"],
        order=balsamic_status_data["order"],
        ordered=dt.datetime.now(),
        ticket=123456,
        cases=balsamic_status_data["families"],
    )

    # THEN it should create and link samples and the family
    assert len(new_families) == 1
    new_family = new_families[0]
    assert new_family.name == "family1"
    assert new_family.data_analysis == str(Pipeline.BALSAMIC)
    assert set(new_family.panels) == set()
    assert new_family.priority_human == "standard"

    assert len(new_family.links) == 1
    new_link = new_family.links[0]
    assert new_link.sample.name == "s1"
    assert new_link.sample.sex == "male"
    assert new_link.sample.application_version.application.tag == "WGTPCFC030"
    assert new_link.sample.comment == "comment"
    assert new_link.sample.is_tumour

    assert base_store.deliveries().count() == base_store.samples().count()
    for link in new_family.links:
        assert len(link.sample.deliveries) == 1



"""
    API for compressing files. Functionality to compress FASTQ, decompress SPRING and clean files
"""

import logging
from pathlib import Path

from housekeeper.store import models as housekeeper_models

from cg.apps.hk import HousekeeperAPI
from cg.apps.crunchy import CrunchyAPI
from cg.meta.compress import files
from cg.models import CompressionData, FileData

LOG = logging.getLogger(__name__)


class CompressAPI:
    """API for compressing BAM and FASTQ files"""

    def __init__(self, hk_api: HousekeeperAPI, crunchy_api: CrunchyAPI, dry_run: bool = False):

        self.hk_api = hk_api
        self.crunchy_api = crunchy_api
        self.dry_run = dry_run

    def set_dry_run(self, dry_run: bool):
        """Update dry run"""
        self.dry_run = dry_run
        if self.crunchy_api.dry_run is False:
            self.crunchy_api.set_dry_run(dry_run)

    def get_latest_version(self, bundle_name: str) -> housekeeper_models.Version:
        """Fetch the latest version of a hk bundle"""
        last_version = self.hk_api.last_version(bundle_name)
        if not last_version:
            LOG.warning("No bundle found for %s in housekeeper", bundle_name)
            return None
        LOG.debug("Found version obj for %s: %s", bundle_name, repr(last_version))
        return last_version

    # Compression methods
    def compress_fastq(self, sample_id: str) -> bool:
        """Compress the FASTQ files for a individual"""
        LOG.info("Check if FASTQ compression is possible for %s", sample_id)
        version_obj = self.get_latest_version(sample_id)
        if not version_obj:
            return False

        sample_fastq_dict = files.get_fastq_files(sample_id=sample_id, version_obj=version_obj)
        if not sample_fastq_dict:
            return False

        all_ok = True
        for run_name in sample_fastq_dict:
            LOG.info("Check if compression possible for run %s", run_name)
            compression_object = sample_fastq_dict[run_name]["compression_data"]
            if FileData.is_empty(compression_object.fastq_first):
                LOG.warning(
                    "Fastq files are empty for %s: %s", sample_id, compression_object.fastq_first
                )
                self.delete_fastq_housekeeper(
                    hk_fastq_first=sample_fastq_dict[run_name]["hk_first"],
                    hk_fastq_second=sample_fastq_dict[run_name]["hk_second"],
                )
                all_ok = False
                continue

            if not self.crunchy_api.is_fastq_compression_possible(compression_object):
                LOG.warning("FASTQ to SPRING not possible for %s, run %s", sample_id, run_name)
                all_ok = False
                continue

            LOG.info(
                "Compressing %s and %s for sample %s into SPRING format",
                compression_object.fastq_first,
                compression_object.fastq_second,
                sample_id,
            )
            self.crunchy_api.fastq_to_spring(compression_object, sample_id=sample_id)

        return all_ok

    def decompress_spring(self, sample_id: str) -> bool:
        """Decompress SPRING archive for a sample

        This function will make sure that everything is ready for decompression from SPRING archive
        to FASTQ files.

            - Housekeeper will be updated to include FASTQ files
            - Housekeeper will still have the SPRING and SPRING metadata file
            - The SPRING metadata file will be updated to include date for decompression
        """
        version_obj = self.get_latest_version(sample_id)
        if not version_obj:
            return False

        compression_objs = files.get_spring_paths(version_obj)
        for compression_obj in compression_objs:
            if not self.crunchy_api.is_spring_decompression_possible(compression_obj):
                LOG.info("SPRING to FASTQ decompression not possible for %s", sample_id)
                return False

            LOG.info(
                "Decompressing %s to FASTQ format for sample %s ",
                compression_obj.spring_path,
                sample_id,
            )

            self.crunchy_api.spring_to_fastq(compression_obj, sample_id=sample_id)
            self.crunchy_api.update_metadata_date(compression_obj.spring_metadata_path)

        return True

    def clean_fastq(self, sample_id: str) -> bool:
        """Check that FASTQ compression is completed for a case and clean

        This means removing compressed FASTQ files and update housekeeper to point to the new SPRING
        file and its metadata file
        """
        LOG.info("Clean FASTQ files for %s", sample_id)
        version_obj = self.get_latest_version(sample_id)
        if not version_obj:
            return False

        sample_fastq_dict = files.get_fastq_files(sample_id=sample_id, version_obj=version_obj)
        if not sample_fastq_dict:
            return False

        all_cleaned = True
        for run_name in sample_fastq_dict:

            compression_object = sample_fastq_dict[run_name]["compression_data"]

            if not self.crunchy_api.is_fastq_compression_done(compression_object):
                LOG.info("FASTQ compression not done for sample %s, run %s", sample_id, run_name)
                all_cleaned = False
                continue

            LOG.info("FASTQ compression done for sample %s, run %s!!", sample_id, run_name)

            self.update_fastq_hk(
                sample_id=sample_id,
                compression_obj=compression_object,
                hk_fastq_first=sample_fastq_dict[run_name]["hk_first"],
                hk_fastq_second=sample_fastq_dict[run_name]["hk_second"],
            )

            self.remove_fastq(
                fastq_first=compression_object.fastq_first,
                fastq_second=compression_object.fastq_second,
            )

        if all_cleaned:
            LOG.info("All FASTQ files cleaned for %s!!", sample_id)

        return all_cleaned

    def add_decompressed_fastq(self, sample_id) -> bool:
        """Adds unpacked FASTQ files to housekeeper"""
        LOG.info("Adds FASTQ to Housekeeper for %s", sample_id)
        version_obj = self.get_latest_version(sample_id)
        if not version_obj:
            return False

        spring_paths = files.get_spring_paths(version_obj)
        for compression_object in spring_paths:
            if not self.crunchy_api.is_spring_decompression_done(compression_object):
                LOG.info("SPRING to FASTQ decompression not finished %s", sample_id)
                return False

            fastq_first = compression_object.fastq_first
            fastq_second = compression_object.fastq_second
            if files.is_file_in_version(
                version_obj=version_obj, path=fastq_first
            ) or files.is_file_in_version(version_obj=version_obj, path=fastq_second):
                LOG.warning("FASTQ files already exists in Housekeeper")
                continue

            LOG.info("Adding decompressed FASTQ files to Housekeeper for sample %s ", sample_id)

            self.add_fastq_hk(
                sample_id=sample_id, fastq_first=fastq_first, fastq_second=fastq_second
            )

        return True

    def delete_fastq_housekeeper(
        self, hk_fastq_first: housekeeper_models.File, hk_fastq_second: housekeeper_models.File
    ) -> None:
        """Delete fastq files from housekeeper"""
        LOG.info("Deleting FASTQ files from Housekeeper")
        hk_fastq_first.delete()
        hk_fastq_second.delete()
        self.hk_api.commit()
        LOG.debug("FASTQ files deleted from Housekeeper")

    # Methods to update housekeeper
    def update_fastq_hk(
        self,
        sample_id: str,
        compression_obj: CompressionData,
        hk_fastq_first: housekeeper_models.File,
        hk_fastq_second: housekeeper_models.File,
    ) -> None:
        """Update Housekeeper with compressed FASTQ files and SPRING metadata file"""
        version_obj = self.hk_api.last_version(sample_id)

        spring_tags = [sample_id, "spring"]
        spring_metadata_tags = [sample_id, "spring-metadata"]
        LOG.info("Updating FASTQ files in Housekeeper update for %s:", sample_id)
        LOG.info(
            "%s, %s -> %s, with tags %s",
            compression_obj.fastq_first,
            compression_obj.fastq_second,
            compression_obj.spring_path,
            spring_tags,
        )
        LOG.info(
            "Adds %s, with tags %s", compression_obj.spring_metadata_path, spring_metadata_tags
        )
        if self.dry_run:
            return

        LOG.info("Updating files in Housekeeper...")
        if files.is_file_in_version(version_obj, compression_obj.spring_path):
            LOG.info("SPRING file is already in Housekeeper")
        else:
            LOG.info("Adding SPRING file to Housekeeper")
            self.hk_api.add_file(
                path=compression_obj.spring_path, version_obj=version_obj, tags=spring_tags
            )
            self.hk_api.commit()

        if files.is_file_in_version(version_obj, compression_obj.spring_metadata_path):
            LOG.info("Spring metadata file is already in Housekeeper")
        else:
            self.hk_api.add_file(
                path=compression_obj.spring_metadata_path,
                version_obj=version_obj,
                tags=spring_metadata_tags,
            )
            self.hk_api.commit()

        self.delete_fastq_housekeeper(hk_fastq_first, hk_fastq_second)

    def add_fastq_hk(self, sample_id: str, fastq_first: Path, fastq_second: Path) -> None:
        """Add FASTQ files to housekeeper"""
        fastq_tags = [sample_id, "fastq"]
        last_version = self.hk_api.last_version(bundle=sample_id)
        LOG.info(
            "Adds %s, %s to bundle %s with tags %s",
            fastq_first,
            fastq_second,
            sample_id,
            fastq_tags,
        )
        if self.dry_run:
            return

        LOG.info("Updating files in Housekeeper...")
        self.hk_api.add_file(path=fastq_first, version_obj=last_version, tags=fastq_tags)
        self.hk_api.add_file(path=fastq_second, version_obj=last_version, tags=fastq_tags)
        self.hk_api.commit()

    # Methods to remove files from disc
    def remove_fastq(self, fastq_first: Path, fastq_second: Path):
        """Remove FASTQ files"""
        LOG.info("Will remove %s and %s", fastq_first, fastq_second)
        if self.dry_run:
            return
        fastq_first.unlink()
        LOG.debug("First FASTQ in pair %s removed", fastq_first)
        fastq_second.unlink()
        LOG.debug("Second FASTQ in pair %s removed", fastq_second)
        LOG.info("FASTQ files removed")
